/*--------------------------------------------------------------------
 *	$Id: grdfilter.c,v 1.32 2006/02/06 14:00:37 remko Exp $
 *
 *	Copyright (c) 1991-2006 by P. Wessel and W. H. F. Smith
 *	See COPYING file for copying and redistribution conditions.
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation; version 2 of the License.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	Contact info: gmt.soest.hawaii.edu
 *--------------------------------------------------------------------*/
/*
 * grdfilter.c  reads a grdfile and creates filtered grd file
 *
 * user selects boxcar, gaussian, cosine arch, median, or mode filter
 * user selects distance scaling as appropriate for map work, etc.
 *
 * Author:  W.H.F. Smith
 * Date: 	16 Aug 88
 *
 * Modified:	27 Sep 88 by W.H.F. Smith, to use new median routine.
 *
 * Updated:	PW: 13-Jun-1991 to v2.0
 *		PW: 13-Jul-1992 to actually do v2 i/o.
 *		PW: 15-Jul-1992 to handle arbitrary new -R -I
 *		PW: 03-Jan-1995 to offer mode-filter (from filter1d)
 *		WS: 21-May-1998 to make warnings silent unless -V on.
 *		PW: 03-Jun-1998 upgrade to GMT 3.1
 *		PW: 02-Jun-1999 upgrade to GMT 3.3 + added SINCOS option
 *		PW: 18-Oct-1999 Use sincos directly
 *		PW: 18-JUN-2000 3.3.5
 * Version:	4
 *		PW: 27-MAY-2004 Added extreme values filter options l, L, u, U
 *		PW: 16-NOV-2005 Added spherical filtering support & highpass option
*/

#define GMT_WITH_NO_PS
#include "gmt.h"

#define FILTER_WIDTH		0
#define HALF_FILTER_WIDTH	1
#define X_SCALE			2
#define Y_SCALE			3
#define INV_R_SCALE		4

struct FILTER_INFO {
	int nx;			/* The max number of filter weights in x-direction */
	int ny;			/* The max number of filter weights in y-direction */
	int x_half_width;	/* Number of filter nodes to either side needed at this latitude */
	int y_half_width;	/* Number of filter nodes above/below this point (ny_f/2) */
	int d_flag;
	double dx, dy;		/* Grid spacing in original units */
	double *x, *y;		/* Distances in original units along x and y to distance nodes */
	double par[5];		/* [0] is filter width, [1] is 0.5*filter_width, [2] is xscale, [3] is yscale, [4] is 1/r_half for filter */
	double x_off, y_off;	/* Offsets relative to original grid */
	PFD weight_func, radius_func;
};

double get_wt();
void   set_weight_matrix (struct FILTER_INFO *F, double output_lat, double par[], double x_off, double y_off, BOOLEAN fast);
double UnitWeight (double r, double par[]);
double CosBellWeight (double r, double par[]);
double GaussianWeight (double r, double par[]);
double CartRadius (double x0, double y0, double x1, double y1, double par[]);
double CartScaledRadius (double x0, double y0, double x1, double y1, double par[]);
double FlatEarthRadius (double x0, double y0, double x1, double y1, double par[]);
double SphericalRadius (double x0, double y0, double x1, double y1, double par[]);

int	*i_origin;
float	*input, *output;
double	*weight, *work_array, *x_shift;

char *filter_name[9] = {
	"Boxcar",
	"Cosine Arch",
	"Gaussian",
	"Median",
	"Mode",
	"Lower",
	"Lower+",
	"Upper",
	"Upper-"
};

int GMT_mode_selection = 0, GMT_n_multiples = 0;

int main (int argc, char **argv)
{
	int	nx_out, ny_out, n_in_median, n_nan = 0, i_west, i_east, j_s_pole, j_n_pole;
	int	j_origin, i_out, j_out, half_nx, i_orig = 0;
	int	i_in, j_in, ii, jj, i, j, ij_in, ij_out, ij_wt, effort_level;
	int	distance_flag, filter_type, one_or_zero = 1;

	BOOLEAN	error, new_range, new_increment, fast_way, shift = FALSE, slow = FALSE, toggle = FALSE;
	BOOLEAN wrap_case = FALSE, highpass = FALSE, pole_check = FALSE;
	BOOLEAN full_360, duplicate_check, i_west_used, i_east_used, row_ok;

	double	west_new, east_new, south_new, north_new, dx_new, dy_new, offset;
	double	filter_width, x_scale, y_scale, x_width, y_width, y, par[5], pole_weight;
	double	x_out, y_out, wt_sum, value, last_median, this_median, xincnew2, yincnew2;
	double	xincold2, yincold2, y_shift = 0.0, x_fix = 0.0, y_fix = 0.0;

	char	*fin = CNULL, *fout = CNULL, c;

	struct	GRD_HEADER h, test_h;
	
	struct FILTER_INFO F;

	argc = GMT_begin (argc, argv);

	error = new_range = new_increment = FALSE;
	fin = fout = NULL;
	filter_width = dx_new = dy_new = west_new = east_new = 0.0;
	filter_type = distance_flag = -1;

	for (i = 1; i < argc; i++) {
		if (argv[i][0] == '-') {
			switch (argv[i][1]) {
				/* Common parameters */

				case 'R':
				case 'V':
				case '\0':
					error += GMT_get_common_args (argv[i], &west_new, &east_new, &south_new, &north_new);
					break;

				case 'F':
					switch (argv[i][2]) {
						case 'b':
							filter_type = 0;
							break;
						case 'c':
							filter_type = 1;
							break;
						case 'g':
							filter_type = 2;
							break;
						case 'm':
							filter_type = 3;
							break;
						case 'p':
							filter_type = 4;
							c = argv[i][strlen(argv[i]-1)];
							if (c == '-') GMT_mode_selection = -1;
							if (c == '+') GMT_mode_selection = +1;
							break;
						case 'l':
							filter_type = 5;
							break;
						case 'L':
							filter_type = 6;
							break;
						case 'u':
							filter_type = 7;
							break;
						case 'U':
							filter_type = 8;
							break;
						default:
							error = TRUE;
							break;
					}
					filter_width = atof(&argv[i][3]);
					highpass = (filter_width < 0.0);
					filter_width = fabs (filter_width);
					break;
				case 'G':
					fout = &argv[i][2];
					break;
				case 'D':
					distance_flag = atoi(&argv[i][2]);
					break;
				case 'I':
					if (GMT_getinc (&argv[i][2], &dx_new, &dy_new)) {
						GMT_inc_syntax ('I', 1);
						error = TRUE;
					}
					new_increment = TRUE;
					break;
				case 'N':	/* Pixel registration OBSOLETE but BACKWARD COMPATIBLE */
					one_or_zero = 0;
					break;
				case 'T':	/* Toggle registration */
					toggle = TRUE;
					break;
				default:
					error = TRUE;
					GMT_default_error (argv[i][1]);
					break;
			}
		}
		else
			fin = argv[i];
	}

	if (argc == 1 || GMT_quick) {
		fprintf (stderr, "grdfilter %s - Filter a 2-D netCDF grdfile in the Time domain\n\n", GMT_VERSION);
		fprintf(stderr,"usage: grdfilter input_file -D<distance_flag> -F<type>[-]<filter_width>[<mode>]\n");
		fprintf(stderr,"\t-G<output_file> [-I<xinc[u][!|+]>[/<yinc>[u][!|+]]] [-R<west/east/south/north>] [-T] [-V]\n");

		if (GMT_quick) exit (EXIT_FAILURE);

		fprintf(stderr,"\tDistance flag determines how grid (x,y) maps into distance units of filter width as follows:\n");
		fprintf(stderr,"\t   -D0 grid x,y same units as <filter_width>, cartesian Distances.\n");
		fprintf(stderr,"\t   -D1 grid x,y in degrees, <filter_width> in km, cartesian Distances.\n");
		fprintf(stderr,"\t   -D2 grid x,y in degrees, <filter_width> in km, x_scaled by cos(middle y), cartesian Distances.\n");
		fprintf(stderr,"\t   These first three options are faster; they allow weight matrix to be computed only once.\n");
		fprintf(stderr,"\t   Next two options are slower; weights must be recomputed for each scan line.\n");
		fprintf(stderr,"\t   -D3 grid x,y in degrees, <filter_width> in km, x_scale varies as cos(y), cartesian Distances.\n");
		fprintf(stderr,"\t   -D4 grid x,y in degrees, <filter_width> in km, spherical Distances.\n");
		fprintf(stderr,"\t-F sets the low-pass filter type and full diameter (6 sigma) filter-width.  Choose between\n");
		fprintf(stderr,"\t   convolution-type filters which differ in how weights are assigned and geospatial\n");
		fprintf(stderr,"\t   filters that seek to return a representative value.\n");
		fprintf(stderr,"\t   Give negative filter width to select highpass filtering [lowpass].\n");
		fprintf(stderr,"\t   Convolution filters:\n");
		fprintf(stderr,"\t     b: Boxcar : a simple averaging of all points inside filter radius.\n");
		fprintf(stderr,"\t     c: Cosine arch : a weighted averaging with cosine arc weights\n");
		fprintf(stderr,"\t     g: Gaussian : weighted averaging with Gaussian weights.\n");
		fprintf(stderr,"\t   Geospatial filters:\n");
		fprintf(stderr,"\t     l: Lower : return minimum of all points.\n");
		fprintf(stderr,"\t     L: Lower+ : return minimum of all +ve points.\n");
		fprintf(stderr,"\t     m: Median : return the median value of all points.\n");
		fprintf(stderr,"\t     p: Maximum likelihood probability estimator : return mode of all points.\n");
		fprintf(stderr,"\t        By default, we return the average if more than one mode is found.\n");
		fprintf(stderr,"\t        Append - or + to the width to instead return the smallest or largest mode.\n");
		fprintf(stderr,"\t     u: Upper : return maximum of all points.\n");
		fprintf(stderr,"\t     U: Upper- : return maximum of all -ve points.\n");
		fprintf(stderr,"\t-G sets output name for filtered grdfile\n");
		fprintf(stderr, "\n\tOPTIONS:\n");
		GMT_inc_syntax ('I', 0);
		fprintf(stderr,"\t   The new xinc and yinc should be divisible by the old ones (new lattice is subset of old).\n");
		fprintf(stderr, "\t-T Toggles between grid and pixel registration for output grid [Default is same as input registration]\n");
		fprintf(stderr, "\t-R for new Range of output grid; enter <WESN> (xmin, xmax, ymin, ymax) separated by slashes.\n");
		GMT_explain_option ('V');
		exit (EXIT_FAILURE);
	}

	if (!fout) {
		fprintf (stderr, "%s: GMT SYNTAX ERROR -G option:  Must specify output file\n", GMT_program);
		error++;
	}
	if (!fin) {
		fprintf (stderr, "%s: GMT SYNTAX ERROR:  Must specify input file\n", GMT_program);
		error++;
	}
	if (distance_flag < 0 || distance_flag > 4) {
		fprintf (stderr, "%s: GMT SYNTAX ERROR -D option:  Choose from the range 0-4\n", GMT_program);
		error++;
	}
	if (filter_type < 0 || filter_width <= 0.0) {
		fprintf (stderr, "%s: GMT SYNTAX ERROR -F option:  Correct syntax:\n", GMT_program);
		fprintf (stderr, "\t-FX<width>, with X one of bcgmp, width is filter fullwidth\n");
		error++;
	}
	if (toggle && one_or_zero == 0) {	/* Both -N and -T set, not good */
		fprintf (stderr, "%s: GMT SYNTAX ERROR -T option:  Not allowed with obsolete -N option\n", GMT_program);
		error++;
	}
	if (project_info.region_supplied && new_increment && highpass) {
		fprintf (stderr, "%s: GMT SYNTAX ERROR -F option:  Highpass filtering requires original -R -I\n", GMT_program);
		error++;
	}
	
	if (error) exit (EXIT_FAILURE);

	GMT_put_history (argc, argv);	/* Update .gmtcommands4 */

	if (project_info.region_supplied) new_range = TRUE;

	if (GMT_read_grd_info (fin, &h)) {
		fprintf (stderr, "%s: Error opening file %s\n", GMT_program, fin);
		exit (EXIT_FAILURE);
	}
	GMT_grd_init (&h, argc, argv, TRUE);	/* Update command history only */

	if (toggle) {	/* Make output grid of the opposite registration */
		one_or_zero = (h.node_offset) ? 1 : 0;
	}
	else
		one_or_zero = (h.node_offset) ? 0 : 1;

	/* Read the input grd file and close it  */

	input = (float *) GMT_memory (VNULL, (size_t)(h.nx * h.ny), sizeof(float), GMT_program);

	if (GMT_read_grd (fin, &h, input, 0.0, 0.0, 0.0, 0.0, GMT_pad, FALSE)) {
		fprintf (stderr, "%s: Error reading file %s\n", GMT_program, fin);
		exit (EXIT_FAILURE);
	}

	last_median = 0.5 * (h.z_min + h.z_max);

	/* Check range of output area and set i,j offsets, etc.  */

	if (!new_range) {
		west_new = h.x_min;
		east_new = h.x_max;
		south_new = h.y_min;
		north_new = h.y_max;
	}
	if (!new_increment) {
		dx_new = h.x_inc;
		dy_new = h.y_inc;
	}

	if (west_new < h.x_min) error = TRUE;
	if (east_new > h.x_max) error = TRUE;
	if (south_new < h.y_min) error = TRUE;
	if (north_new > h.y_max) error = TRUE;
	if (dx_new <= 0.0) error = TRUE;
	if (dy_new <= 0.0) error = TRUE;

	if (error) {
		fprintf(stderr,"%s: New WESN incompatible with old.\n", GMT_program);
		exit (EXIT_FAILURE);
	}

	/* Make sure output grid is kosher */

	test_h.x_min = west_new;	test_h.x_max = east_new;	test_h.x_inc = dx_new;
	test_h.y_min = south_new;	test_h.y_max = north_new;	test_h.y_inc = dy_new;
	GMT_RI_prepare (&test_h);	/* Ensure -R -I consistency and set nx, ny */
	GMT_grd_RI_verify (&test_h, 1);

	/* We can save time by computing a weight matrix once [or once pr scanline] only
	   if new grid spacing is a multiple of old spacing */

	fast_way = (fabs (fmod (dx_new / h.x_inc, 1.0)) < SMALL && fabs (fmod (dy_new / h.y_inc, 1.0)) < SMALL);

	if (!fast_way && gmtdefs.verbose) {
		fprintf (stderr, "%s: Warning - Your output grid spacing is such that filter-weights must\n", GMT_program);
		fprintf (stderr, "be recomputed for every output node, so expect this run to be slow.  Calculations\n");
		fprintf (stderr, "can be speeded up significantly if output grid spacing is chosen to be a multiple\n");
		fprintf (stderr, "of the input grid spacing.  If the odd output grid is necessary, consider using\n");
		fprintf (stderr, "a \'fast\' grid for filtering and then resample onto your desired grid with grdsample.\n");
	}

	nx_out = one_or_zero + irint ( (east_new - west_new) / dx_new);
	ny_out = one_or_zero + irint ( (north_new - south_new) / dy_new);

	output = (float *) GMT_memory (VNULL, (size_t)(nx_out*ny_out), sizeof(float), GMT_program);
	i_origin = (int *) GMT_memory (VNULL, (size_t)nx_out, sizeof(int), GMT_program);
	if (!fast_way) x_shift = (double *) GMT_memory (VNULL, (size_t)nx_out, sizeof(double), GMT_program);

	xincnew2 = (one_or_zero) ? 0.0 : 0.5 * dx_new;
	yincnew2 = (one_or_zero) ? 0.0 : 0.5 * dy_new;
	xincold2 = (h.node_offset) ? 0.5 * h.x_inc : 0.0;
	yincold2 = (h.node_offset) ? 0.5 * h.y_inc : 0.0;
	offset = (h.node_offset) ? 0.5 : 0.0;

	if (fast_way && h.node_offset == one_or_zero) {	/* multiple grid but one is pix, other is grid */
		x_fix = 0.5 * h.x_inc;
		y_fix = 0.5 * h.y_inc;
		shift = (x_fix != 0.0 || y_fix != 0.0);
	}
	full_360 = (distance_flag && fabs (fabs (h.x_max - h.x_min) - 360.0) < GMT_CONV_LIMIT);	/* Periodic geographic grid */

	/* Set up the distance scalings for lon and lat, and assign pointer to distance function  */

	switch (distance_flag) {
		case 0:	/* Plain, unscaled isotropic Cartesian distances */
			x_scale = y_scale = 1.0;
			F.radius_func = CartRadius;
			break;
		case 1:	/* Plain, scaled (degree to km) isotropic Cartesian distances */
			x_scale = y_scale = project_info.KM_PR_DEG;
			F.radius_func = CartScaledRadius;
			break;
		case 2:	/* Flat Earth Cartesian distances, xscale fixed at mid latitude */
			x_scale = project_info.KM_PR_DEG * cosd (0.5 * (north_new + south_new));
			y_scale = project_info.KM_PR_DEG;
			F.radius_func = FlatEarthRadius;
			break;
		case 3:	/* Flat Earth Cartesian distances, xscale reset for each latitude */
			x_scale = project_info.KM_PR_DEG * ((fabs (south_new) > north_new) ? cosd (south_new) : cosd (north_new));
			y_scale = project_info.KM_PR_DEG;
			F.radius_func = FlatEarthRadius;
			break;
		default:	/* Great circle distances */
			x_scale = 0.0;
			y_scale = project_info.KM_PR_DEG;
			F.radius_func = SphericalRadius;
			if (full_360) wrap_case = TRUE;	/* For spherical filtering */
			break;
	}
			
	switch (filter_type) {
		case 1:	/*  Cosine-bell filter weights */
			par[INV_R_SCALE] = 2.0 / filter_width;
			F.weight_func = CosBellWeight;
			break;
		case 2:	/*  Gaussian filter weights */
			par[INV_R_SCALE] = -18.0 / (filter_width * filter_width);
			F.weight_func = GaussianWeight;
			break;
		default:	/* Everything else uses unit weights */
			F.weight_func = UnitWeight;
			break;
	}
	
	/* Set up miscellaneous filter parameters needed when computing the weights */
	
	par[FILTER_WIDTH] = filter_width;
	par[HALF_FILTER_WIDTH] = 0.5 * filter_width;
	par[X_SCALE] = x_scale;
	par[Y_SCALE] = y_scale;
	F.d_flag = distance_flag;
	F.dx = h.x_inc;
	F.dy = h.y_inc;
	x_width = filter_width / (h.x_inc * x_scale);
	y_width = filter_width / (h.y_inc * y_scale);
	F.y_half_width = (int) (ceil(y_width) / 2.0);
	F.x_half_width = (int) (ceil(x_width) / 2.0);
	F.nx = 2 * F.x_half_width + 1;
	F.ny = 2 * F.y_half_width + 1;
	if (x_scale == 0.0 || F.nx < 0 || F.nx > h.nx) {	/* Safety valve when x_scale -> 0.0 */
		F.nx = h.nx + 1;
		F.x_half_width = h.nx / 2;
	}
	if (F.ny < 0 || F.ny > h.ny) {
		F.ny = h.ny;
		F.y_half_width = h.ny / 2;
	}
	F.x = (double *) GMT_memory (VNULL, (size_t)(F.x_half_width+1), sizeof (double), GMT_program);
	F.y = (double *) GMT_memory (VNULL, (size_t)(F.y_half_width+1), sizeof (double), GMT_program);
	for (i = 0; i <= F.x_half_width; i++) F.x[i] = i * F.dx;
	for (j = 0; j <= F.y_half_width; j++) F.y[j] = j * F.dy;
	
	weight = (double *) GMT_memory (VNULL, (size_t)(F.nx*F.ny), sizeof(double), GMT_program);

	if (filter_type >= 3) {	/* These filters are not convolutions; they require sorting or comparisons */
		slow = TRUE;
		work_array = (double *) GMT_memory (VNULL, (size_t)(F.nx*F.ny), sizeof(double), GMT_program);
	}

	if (gmtdefs.verbose) {
		fprintf(stderr,"%s: Input nx,ny = (%d %d), output nx,ny = (%d %d), filter nx,ny = (%d %d)\n", GMT_program, h.nx, h.ny, nx_out, ny_out, F.nx, F.ny);
		fprintf(stderr,"%s: Filter type is %s.\n", GMT_program, filter_name[filter_type]);
	}

	/* Compute nearest xoutput i-indices and shifts once */

	for (i_out = 0; i_out < nx_out; i_out++) {
		x_out = west_new + i_out * dx_new + xincnew2;
		i_origin[i_out] = GMT_x_to_i (x_out, h.x_min, h.x_inc, h.xy_off, h.nx);
		if (!fast_way) x_shift[i_out] = x_out - (h.x_min + i_origin[i_out] * h.x_inc + xincold2);
	}

	/* Determine how much effort to compute weights:
		1 = Compute weights once for entire grid
		2 = Compute weights once per scanline
		3 = Compute weights for every output point [slow]
	*/

	if (fast_way && distance_flag <= 2)
		effort_level = 1;
	else if (fast_way && distance_flag > 2)
		effort_level = 2;
	else 
		effort_level = 3;
	
	if (effort_level == 1) set_weight_matrix (&F, 0.0, par, x_fix, y_fix, shift);
	half_nx = (h.node_offset) ? h.nx / 2 : (h.nx - 1) / 2;
	pole_weight = 0.25 * M_PI * (h.y_inc / h.x_inc);	/* This is the weight of the pole point for gridregistered grids */
	pole_check = (distance_flag && !h.node_offset);		/* Must check to make sure we only use the N and S pole value once */
	duplicate_check = (full_360 && !h.node_offset);		/* Must avoid using the duplicated Greenwhich node twice */
	j_s_pole = (h.y_min == -90.0) ? h.ny - 1 : INT_MIN;	/* row number of S pole (if in the grid)*/
	j_n_pole = (h.y_max == 90.0) ? 0 : INT_MIN;		/* row number of N pole (if in the grid)*/
	i_west = 0;		/* column number of western edge */
	i_east = h.nx - 1;	/* column number of eastern edge */
	
	for (j_out = 0; j_out < ny_out; j_out++) {

		if (gmtdefs.verbose) fprintf (stderr, "%s: Processing output line %d\r", GMT_program, j_out);
		y_out = north_new - j_out * dy_new - yincnew2;
		j_origin = GMT_y_to_j (y_out, h.y_min, h.y_inc, h.xy_off, h.ny);
		if (distance_flag == 3) par[X_SCALE] = project_info.KM_PR_DEG * cos (D2R * y_out);	/* Update flat-earth longitude scale */

		if (distance_flag > 2) {	/* Update max filterweight nodes to deal with at this latitude */
			y = fabs (y_out);
			if (distance_flag == 4) y += (par[HALF_FILTER_WIDTH] / par[Y_SCALE]);
			F.x_half_width = (y < 90.0) ? MIN (F.nx / 2, irint (par[HALF_FILTER_WIDTH] / (F.dx * par[Y_SCALE] * cosd (y)))) : F.nx / 2;
		}
			
		if (effort_level == 2) set_weight_matrix (&F, y_out, par, x_fix, y_fix, shift);
		if (!fast_way) y_shift = y_out - (h.y_max - j_origin * h.y_inc - yincold2);

		for (i_out = 0; i_out < nx_out; i_out++) {

			if (effort_level == 3) set_weight_matrix (&F, y_out, par, x_shift[i_out], y_shift, fast_way);
			wt_sum = value = 0.0;
			n_in_median = 0;
			ij_out = j_out * nx_out + i_out;

			/* Now loop over the filter domain and collect those points that should be considered by the filter operation */
			
			i_west_used = i_east_used = FALSE;
			for (ii = -F.x_half_width; ii <= F.x_half_width; ii++) {
				i_in = i_origin[i_out] + ii;
				if (wrap_case) {	/* Just wrap around the globe */
					if (i_in < 0) i_in += h.nx;
					if (i_in >= h.nx) i_in -= h.nx;
					i_orig = i_in;
				}
				else if ( (i_in < 0) || (i_in >= h.nx)) continue;

				if (i_in == i_west) i_west_used = TRUE;
				if (i_in == i_east) i_east_used = TRUE;
				if (duplicate_check && ((i_in == i_east && i_west_used) || (i_in == i_west && i_east_used))) continue;	/* Do not use the node at 360 if the one at 0 was used, or vice versa */
				
				for (jj = -F.y_half_width, row_ok = TRUE; row_ok && jj <= F.y_half_width; jj++) {
					j_in = j_origin + jj;
					
					if (pole_check && (j_in == j_n_pole || j_in == j_s_pole)) {	/* N or S pole on gridline grid */
						row_ok = FALSE;	/* Only use one point from this row since we are at a pole */
						ij_in = j_in*h.nx;	/* All points on pole rows are the same; pick the first one */
						if (GMT_is_fnan (input[ij_in])) continue;	/* Whaddaya know; the pole is NaN */
						/* Get here when pole value is usable  */
						if (slow) {
							work_array[n_in_median] = input[ij_in];
							n_in_median++;
						}
						else {
							value += input[ij_in] * pole_weight;
							wt_sum += pole_weight;
						}
						continue;	/* Done with this row */
					}
						
					if (wrap_case) {	/* Just wrap over the pole */
						i_in = i_orig;
						if (j_in < 0) j_in = -j_in, i_in += half_nx;
						if (j_in >= h.ny) j_in = 2*h.ny - j_in - 2, i_in += half_nx;
						if (i_in < 0) i_in += h.nx;
						if (i_in >= h.nx) i_in -= h.nx;
					}
					else if ( (j_in < 0) || (j_in >= h.ny) ) continue;

					ij_wt = (jj + F.y_half_width) * F.nx + ii + F.x_half_width;
					if (weight[ij_wt] <= 0.0) continue;

					ij_in = j_in*h.nx + i_in;
					if (GMT_is_fnan (input[ij_in])) continue;

					/* Get here when point is usable  */
					if (slow) {
						work_array[n_in_median] = input[ij_in];
						n_in_median++;
					}
					else {
						value += input[ij_in] * weight[ij_wt];
						wt_sum += weight[ij_wt];
					}
				}
			}

			/* Now we have done the convolution and we can get the value  */

			if (slow) {
				if (n_in_median) {
					switch (filter_type) {
						case 3:	/* Median */
							GMT_median (work_array, n_in_median, h.z_min, h.z_max, last_median, &this_median);
							last_median = this_median;
							break;
						case 4:	/* Mode */
							GMT_mode (work_array, n_in_median, n_in_median/2, TRUE, GMT_mode_selection, &GMT_n_multiples, &this_median);
							break;
						case 5:	/* Lowest of all */
							this_median = GMT_extreme (work_array, n_in_median, DBL_MAX, 0, -1);
							break;
						case 6:	/* Lowest of positive values */
							this_median = GMT_extreme (work_array, n_in_median, 0.0, +1, -1);
							break;
						case 7:	/* Upper of all values */
							this_median = GMT_extreme (work_array, n_in_median, -DBL_MAX, 0, +1);
							break;
						case 8:	/* Upper of negative values */
							this_median = GMT_extreme (work_array, n_in_median, 0.0, -1, +1);
							break;
					}
					output[ij_out] = (float)this_median;
				}
				else {
					output[ij_out] = GMT_f_NaN;
					n_nan++;
				}
			}
			else {
				if (wt_sum == 0.0) {	/* Assign value = GMT_f_NaN */
					n_nan++;
					output[ij_out] = GMT_f_NaN;
				}
				else
					output[ij_out] = (float)(value / wt_sum);
			}
		}
	}
	if (gmtdefs.verbose) fprintf (stderr, "\n");

	if (highpass) {
		if (gmtdefs.verbose) fprintf (stderr, "%s: Subtracting lowpass-filtered data from grid to obtain high-pass filtered data\n", GMT_program);
		for (ij_out = 0; ij_out < nx_out * ny_out; ij_out++) output[ij_out] = input[ij_out] - output[ij_out];
	}
	
	/* At last, that's it!  Output: */

	if (n_nan && gmtdefs.verbose) fprintf (stderr, "%s: Unable to estimate value at %d nodes, set to NaN\n", GMT_program, n_nan);
	if (GMT_n_multiples > 0 && gmtdefs.verbose) fprintf (stderr, "%s: WARNING: %d multiple modes found\n", GMT_program, GMT_n_multiples);

	h.nx = nx_out;
	h.ny = ny_out;
	h.x_min = west_new;
	h.x_max = east_new;
	h.y_min = south_new;
	h.y_max = north_new;
	h.x_inc = dx_new;
	h.y_inc = dy_new;
	h.node_offset = !one_or_zero;

	if (GMT_write_grd (fout, &h, output, 0.0, 0.0, 0.0, 0.0, GMT_pad, FALSE)) {
		fprintf (stderr, "%s: Error writing file %s\n", GMT_program, fout);
		exit (EXIT_FAILURE);
	}

	GMT_free ((void *) input);
	GMT_free ((void *) output);
	GMT_free ((void *) weight);
	GMT_free ((void *) i_origin);
	GMT_free ((void *) F.x);
	GMT_free ((void *) F.y);
	if (slow) GMT_free ((void *) work_array);
	if (!fast_way) GMT_free ((void *) x_shift);

	GMT_end (argc, argv);

	exit (EXIT_SUCCESS);
}

void set_weight_matrix (struct FILTER_INFO *F, double output_lat, double par[], double x_off, double y_off, BOOLEAN fast)
{
	/* x_off and y_off give offset between output node and 'origin' input node for this window (0,0 for integral grids).
	 * fast is TRUE when input/output grids are offset by integer values in dx/dy.
	 * Here, par[0] = filter_width, par[1] = filter_width / 2, par[2] = x_scale, part[3] = y_scale, and
	 * par[4] is the normalization distance needed for the Cosine-bell or Gaussian weight function.
	 */

	int	i, j, ij;
	double	x, y, yc, x_jump = 0.0, normalize = 1.0, r;

	yc = output_lat - y_off;
	for (j = -F->y_half_width; j <= F->y_half_width; j++) {
		y = yc + ((j < 0) ? -F->y[-j] : F->y[j]);
		if (F->d_flag) {	/* Must normalize based on shrinking area representation and check for going over the pole */
			if (fabs (y) > 90.0) {	/* Must find the point across the pole */
				y = copysign (180.0 - fabs (y), y);
				x_jump = 180.0;
			}
			else
				x_jump = 0.0;
			normalize = cosd (y);
		}
		for (i = -F->x_half_width; i <= F->x_half_width; i++) {
			x = (i < 0) ? -F->x[-i] : F->x[i];
			ij = (j + F->y_half_width) * F->nx + i + F->x_half_width;
			r = F->radius_func (x_off, output_lat, x + x_jump, y, par);
			weight[ij] = (r > par[HALF_FILTER_WIDTH]) ? -1.0 : F->weight_func (r, par);
			if (F->d_flag) weight[ij] *= normalize;	/* Adjust for variation in area with latitude */
		}
	}
}

double CartRadius (double x0, double y0, double x1, double y1, double par[])
{	/* Plain Cartesian distance */
	return (hypot (x0 - x1, y0 - y1));
}

double CartScaledRadius (double x0, double y0, double x1, double y1, double par[])
{	/* Plain scaled Cartesian distance (xscale = yscale) */
	return (par[X_SCALE] * hypot (x0 - x1, y0 - y1));
}

double FlatEarthRadius (double x0, double y0, double x1, double y1, double par[])
{	/* Cartesian radius with different scales */
	return (hypot (par[X_SCALE] * (x0 - x1), par[Y_SCALE] * (y0 - y1)));
}

double SphericalRadius (double x0, double y0, double x1, double y1, double par[])
{	/* Great circle distance with polar wrap-around test on 2nd point */
	if (fabs (y1) > 90.0) {	/* Must find the point across the pole */
		y1 = copysign (180.0 - fabs (y1), y1);
		x1 += 180.0;
	}
	return (GMT_great_circle_dist_km (x0, y0, x1, y1));
}

double UnitWeight (double r, double par[])
{
	/* Return unit weight since we know r is inside radius */

	return (1.0);
}

double CosBellWeight (double r, double par[])
{
	/* Return the cosine-bell filter weight for given r.
	 * The parameter r_f_half is the 5th parameter passed.
	 */

	return (1.0 + cos (M_PI * r * par[INV_R_SCALE]));
}

double GaussianWeight (double r, double par[])
{
	/* Return the Gaussian filter weight for given r.
	 * The parameter sig_2 is the 5th parameter passed.
	 */

	return (exp (r * r * par[INV_R_SCALE]));
}
